"""
This module provides 3D neural networks for DSCNet, UNet, and VNet

Copyright (c) 2024, Auorui.
All rights reserved.
time 2024-04-28 The last weekend of April.
"""
from .basic_unet import BasicUNet3D
from .vent import VNet3D
from .dscnet import DSCNet3D